/* 
 *	HT Editor
 *	htapp.cc
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "htapp.h"
#include "htclipboard.h"
#include "htcolor.h"
#include "htdialog.h"
#include "hthelp.h"
#include "hthist.h"
#include "htiobox.h"
#include "htkeyb.h"
#include "htmenu.h"
#include "htpal.h"
#include "htsearch.h"
#include "htstream.h"
#include "htstring.h"
#include "httree.h"
#include "tools.h"

#include "viewers.h"

#include <errno.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

#define keyline_palette_str			"keyline"
#define keyline_palette_colors          3
#define keyline_coloridx_digit		0
#define keyline_coloridx_active		1
#define keyline_coloridx_inactive		2
#define keyline_color_digit			pal[keyline_coloridx_digit].value
#define keyline_color_active			pal[keyline_coloridx_active].value
#define keyline_color_inactive		pal[keyline_coloridx_inactive].value

#define desktop_palette_str				"desktop"
#define desktop_palette_colors			1
#define desktop_coloridx_body  			0
#define desktop_color_body  				pal[desktop_coloridx_body].value

char _globalerror[1024];
char *globalerror=_globalerror;

/*
 *	CLASS ht_keyline
 */

void ht_keyline::init(bounds *b)
{
	ht_view::init(b, 0, 0);
	VIEW_DEBUG_NAME("ht_keyline");
}

void ht_keyline::done()
{
	ht_view::done();
}

void ht_keyline::defaultpalette()
{
	pal_name=keyline_palette_str;
}

void ht_keyline::draw()
{
	clear(keyline_color_inactive);
	int x=0;
	for (int i=1; i<=10; i++) {
		htmsg msg;
		msg.type=htmt_nothing;
		msg.msg=htm_funcname;
		msg.data_int=i;
		baseview->sendmsg(&msg);
		printf(x, 0, keyline_color_digit, "%d", i%10);
		if (msg.msg==htm_funcname_return) {
			char *s=msg.data_str;
			if (s) {
				if (s[0]=='~') {
					printf(x+1, 0, keyline_color_inactive, s+1);
				} else {
					for (int j=0; j<size.w/10-1; j++) {
						printf(x+j+1, 0, keyline_color_active, " ");
					}
					printf(x+1, 0, keyline_color_active, s);
				}
			}
		}
		x+=size.w/10;
	}
}

void ht_keyline::handlemsg(htmsg *msg)
{
	ht_view::handlemsg(msg);
}

/*
 *	CLASS ht_desktop
 */

void ht_desktop::init(bounds *b)
{
	ht_view::init(b, VO_OWNBUFFER, 0);
	VIEW_DEBUG_NAME("ht_desktop");
}

void ht_desktop::done()
{
	ht_view::done();
}

void ht_desktop::defaultpalette()
{
	pal_name=desktop_palette_str;
}

void ht_desktop::draw()
{
	fill(0, 0, size.w, size.h, desktop_color_body, CHAR_FILLED_M);
}

/*
 *   app_stream_error_proc()
 */

int app_stream_error_proc(ht_stream *stream)
{
	int err=stream->get_error();
	char *name=stream->get_name();
	if (err & STERR_SYSTEM) {
		err=err&0xffff;
		switch (err) {
			case 4: {	/* EACCES*/
#ifdef DJGPP
				struct stat sbuf;

				stat(name, &sbuf);

				if (!(sbuf.st_mode & S_IWUSR)) {
					if (msgbox(msgbox_yes|msgbox_no, "title", 1, align_center, "%s: stream error (Permission denied), seems to be a (DOS) read-only file. Change attribute ?", name)==button_yes) {
						if (chmod(name, S_IRUSR | S_IWUSR)) {
							errorbox_modal("%s: error (%04x) changing attribute", name, errno & 0xffff);
						} else {
							stat(name, &sbuf);

							if (!(sbuf.st_mode & S_IWUSR)) {
								errorbox_modal("%s: error (%04x) changing attribute", name, errno & 0xffff);
							} else {
								return SERR_RETRY;
							}
						}
					}
				}

				break;
#endif
			}
			default:
				errorbox_modal("%s: stream error %04x", name, err);
		}
	} else {
		err=err&0xffff;
		errorbox_modal("%s: internal stream error %04x", name, err);
	}
	return SERR_FAIL;
}

/*
 *	CLASS ht_app
 */

/*debug*/
//#define DRAW_TIMINGS
//#define NO_AVG
#define AVG_TIMINGS 10
int timings[AVG_TIMINGS];
int cur_timing=0, max_timing=0;
int h0;
/**/

void ht_app::init(bounds *_b)
{
	ht_group::init(_b, 0, 0);
	VIEW_DEBUG_NAME("ht_app");
	exit_program=0;
	focused=1;
	bounds b;
/* init timer */
	h0=new_timer();
/* create menu */
	getbounds(&b);
	b.x=0;
	b.y=0;
	b.h=1;
	ht_menu *m=new ht_menu();
	m->init(&b);

	ht_static_context_menu *file=new ht_static_context_menu();
	file->init("~File");
	file->insert_entry("~Load", 0, cmd_load, 0, 1);
	file->insert_entry("~Save", 0, 0, 0, 0);
	file->insert_entry("~Save As...", 0, 0, 0, 0);
	file->insert_separator();
	file->insert_entry("~Quit", "Alt-X", cmd_quit, 0, 1);
	m->insert_menu(file);

	ht_static_context_menu *edit=new ht_static_context_menu();
	edit->init("~Clipboard");
	edit->insert_entry("~Copy", "Ctrl+Ins", cmd_clipboard_copy, 0, 1);
	edit->insert_entry("~Paste", "Shift+Ins", cmd_clipboard_paste, 0, 1);
	edit->insert_entry("S~how", 0, cmd_clipboard_show, 0, 1);
	edit->insert_entry("Cl~ear", 0, cmd_clipboard_clear, 0, 1);
	edit->insert_separator();
	edit->insert_entry("Copy ~from file...", 0, cmd_clipboard_copy_from_file, 0, 1);
	edit->insert_entry("Paste ~into file...", 0, cmd_clipboard_paste_into_file, 0, 1);
	m->insert_menu(edit);

	ht_static_context_menu *options=new ht_static_context_menu();
	options->init("~Options");
	options->insert_entry("~Colors", "Ctrl+K", cmd_options_colors, K_Control_K, 1);
//	options->insert_separator();
	m->insert_menu(options);

	ht_static_context_menu *help=new ht_static_context_menu();
	help->init("~Help");
	help->insert_entry("~Help", "F1", cmd_help, 0, 1);
	m->insert_menu(help);

	menu=m;
	insert(menu);
/* create desktop */
	getbounds(&b);
	b.x=0;
	b.y=1;
	b.h-=2;
	desktop=new ht_desktop();
	desktop->init(&b);
	insert(desktop);
/* create keyline */
	getbounds(&b);
	b.x=0;
	b.y=b.h-1;
	b.h=1;
	keyline=new ht_keyline();
	keyline->init(&b);
	insert(keyline);
/* create battlefield */
	getbounds(&b);
	b.x=0;
	b.y=1;
	b.h-=2;

/*     b.x+=5;
	b.y+=2;
	b.w-=10;
	b.h-=4;*/

	battlefield=new ht_group();
	battlefield->init(&b, 0, "battlefield");
	insert(battlefield);
/* create logwindow */
	battlefield->getbounds(&b);
	b.x=0;
	b.y=0;

	logwindow=new ht_window();
	logwindow->init(&b, "log window", 1, FS_KILLER | FS_TITLE);
	b.x=0;
	b.y=0;
	b.w-=2;
	b.h-=2;
	logviewer=new ht_logviewer();
	logviewer->init(&b, 128);
	logwindow->insert(logviewer);
	battlefield->insert(logwindow);
	focus(logwindow);
}

void ht_app::done()
{
	delete_timer(h0);
	ht_group::done();
}

ht_window *ht_app::create_window(char *filename)
{
	bounds b;
	battlefield->getbounds(&b);
	b.x=0;
	b.y=0;
	char fullfilename[FILENAME_MAX];
	canonicalize(filename, fullfilename);
	LOG("loading file %s...", fullfilename);
	ht_file_stream *emfile=new ht_file_stream();
	emfile->init(fullfilename, SAM_READ);
	ht_cached_stream_pbd *file=new ht_cached_stream_pbd();
	file->init(emfile, 1, 256*1024, CACHED_STREAM_CACHE_READS |
		CACHED_STREAM_CACHE_WRITES | CACHED_STREAM_ONLY_EXPLICIT_FLUSHES,
		4096);
	if (file->get_error()) {
		LOG("error loading file %s", fullfilename);
		return 0;
	}
	file->set_error_proc(app_stream_error_proc);
	int winnumber=findfreewinnumber();
	ht_window *window=new ht_window();
	window->init(&b, fullfilename, winnumber, FS_KILLER | FS_TITLE);
	LOG("%s: allocated window %d.", file->get_name(), winnumber);
	b.x=0;
	b.y=0;
	b.w-=2;
	b.h-=2;
	ht_viewergroup *viewergroup=new ht_viewergroup();
	viewergroup->init(&b, VO_SELECTABLE | VO_RESIZE, VIEWERGROUP_NAME, file, viewer_ifs);

	window->insert(viewergroup);
	LOG("%s: done.", fullfilename);
	battlefield->insert(window);
	focus(window);
	return window;
}

int ht_app::load(ht_stream *f)
{
	if (!pal_grp->load(f)) return 0;
	return ht_group::load(f);
}

void ht_app::store(ht_stream *f)
{
	pal_grp->store(f);
	ht_group::store(f);
}

void ht_app::draw()
{
/* show draw timings */
#ifdef DRAW_TIMING
	int xyz=get_timer_1024tick(h0);
	buf->printf(17, 1, 7, "cur: %d (%d msec)", xyz*1024, get_timer_msec(h0));
#ifndef NO_AVG
	if (cur_timing>=AVG_TIMINGS-1) cur_timing=0;
	timings[cur_timing++]=xyz;
	if (cur_timing>max_timing) max_timing=cur_timing;
	int avg=0;
	for (int i=0; i<max_timing; i++) avg+=timings[i];
	avg=avg/max_timing;
	buf->printf(57, 1, 7, "avg%d: %d", max_timing+1, avg*1024);
#endif
#endif
/* display flag if analyser(s) active */
	if (some_analyser_active) buf->printf(60, 0, VCP(VC_TRANSPARENT, VC_TRANSPARENT), "ANALY(%d)", num_ops_parsed);
}

void ht_app::delete_window(ht_window *window)
{
	battlefield->remove(window);
	window->done();
	delete window;
}

int ht_app::findfreewinnumber()
{
	int i=1;
	int free;
	while (1) {
		ht_view *w=battlefield->first;
		free=1;
		while (w) {
			if (i==w->getnumber()) {
				free=0;
				break;
			}
			w=w->next;
		}
		if (free) return i;
		i++;
	}
	return -1;
}

ht_window *ht_app::findwindowbynumber(int index)
{
	ht_view *w=battlefield->first;
	while (w) {
		if (index==w->getnumber()) return (ht_window*)w;
		w=w->next;
	}
	return 0;
}

void ht_app::focus_window(int index)
{
	ht_window *w=findwindowbynumber(index);
	if (w) focus(w);
}

char *ht_app::func(int i, int getname)
{
	switch (i) {
		case 1:
			if (getname) return "help"; else sendmsg(htm_command, cmd_help);
			break;
		case 10:
			if (getname) return "quit"; else sendmsg(htm_command, cmd_quit);
			break;
	}
	return 0;
}

void ht_app::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_draw) {
		start_timer(h0);
		if (msg->type==htmt_broadcast) {
			ht_view *v=first;
			while (v) {
				v->handlemsg(msg);
				v=v->next;
			}
		} else {
			current->handlemsg(msg);
		}
		stop_timer(h0);
		draw();
		screen->show();
	} else {
		ht_group::handlemsg(msg);
	}
	if (msg->msg==htm_keypressed) {
		int i=0;
		switch (msg->data_int) {
			case K_Alt_9: i++;
			case K_Alt_8: i++;
			case K_Alt_7: i++;
			case K_Alt_6: i++;
			case K_Alt_5: i++;
			case K_Alt_4: i++;
			case K_Alt_3: i++;
			case K_Alt_2: i++;
			case K_Alt_1: i++;
				focus_window(i);
				clearmsg(msg);
				break;
			case K_Alt_0:
				popup_window_list();
				clearmsg(msg);
				break;
			case K_Alt_F3:
				sendmsg(htm_kill, battlefield->current);
				break;
			case K_Return:
				popup_view_list();
				clearmsg(msg);
				break;
			case K_Space: {
/*				msgbox(0, "title", 0, align_custom,
"aaa bbb ccc ddd eee fff ggg hhh iii jjj kkk lll mmm nnn ooo ppp
qqq rrr sss ttt uuu vvv www xxx yyy zzz");*/
				msgbox(0, "title", 0, align_custom, "test, baby !\n
Was willst DU Ich komm   nicht dahinter Ich zaehm dir das  Feuer      Du schenkst mir \
nur Winter Wenn du nicht bald anfngst mir Antwort zu geben faengt es ohne \
dich an das richtige Leben Ich hab lange gewartet bin nicht durchgeknallt \
aber wenn du jetzt nicht \n\n\ecaufspringst wirst du ohne mich alt Erklaer mir ich \
warte Wohin geht die Reise \n\n\erich kenn doch die Karte Nimm mich nicht als \
Vorwand nimm mich nicht auf die Schippe sonst kuess ich dich luftleer und \
zerbeiss dir die Lippe ich kann ohne dich leben auch wenn du nicht glaubst \
da ist mehr an mir dran als was du mir raubst ");
			}
		}
	} else if (msg->msg==htm_command) {
		switch (msg->data_int) {
			case cmd_quit:
				exit_program=1;
				clearmsg(msg);
				return;
			case cmd_load: {
				char filename[129];
				filename[0]=0;
				if (inputbox("load file", "filename", filename, 128)==button_ok) {
					create_window(filename);
				}
				clearmsg(msg);
				return;
			}
			case cmd_clipboard_show: {
				popup_clipboard_viewer();
				clearmsg(msg);
				return;
			}
			case cmd_clipboard_clear: {
				if (confirmbox("Do you really want to delete the clipboard ?")==button_ok) {
					clipboard_clear();
				}
				clearmsg(msg);
				return;
			}
			case cmd_clipboard_paste_into_file: {
				char filename[129];
				filename[0]=0;
				if (inputbox("clipboard - paste into file", "filename", filename, 128)==button_ok) {
					ht_file_stream *f=new ht_file_stream();
					f->init(filename, SAM_WRITE | SAM_CREATE);
					if (!f->get_error()) {
						clipboard_paste(f, 0);
					} else errorbox("can't open file '%s'", filename);
					f->done();
					delete f;
				}
				clearmsg(msg);
				return;
			}
			case cmd_clipboard_copy_from_file: {
				char filename[129];
				filename[0]=0;
				if (inputbox("clipboard - copy from file", "filename", filename, 128)==button_ok) {
					char desc[1024]; /* FIXME: possible buffer overflow */
					ht_file_stream *f=new ht_file_stream();
					f->init(filename, SAM_READ);
					if (!f->get_error()) {
						sprintf(desc, "file %s", f->get_name());
						clipboard_copy(desc, f, 0, f->get_size());
					} else errorbox("can't open file '%s'", filename);
					f->done();
					delete f;
				}
				clearmsg(msg);
				return;
			}
			case cmd_options_colors:
				popup_color_list();
				clearmsg(msg);
				return;
			case cmd_help:
				popup_help_dialog("htinfo", "Top");
				clearmsg(msg);
				return;
		}
	} else if (msg->msg==htm_kill) {
		if ((ht_window*)msg->data_ptr!=logwindow) {
			delete_window((ht_window*)msg->data_ptr);
			clearmsg(msg);
		}
	}
}

int ht_app::logf(const char *format, ...)
{
	char buf[1024];	/* FIXME: possible buffer overflow ! */
	va_list arg;
	va_start(arg, format);
	int r=vsprintf(buf, format, arg);
	va_end(arg);
	logviewer->insertline(buf);
	sendmsg(htm_draw, 0);
	return r;
}

void	ht_app::popup_clipboard_viewer()
{
	bounds b;
	b.w=screen->size.w-4;
	b.h=screen->size.h*7/8;
	b.x=(screen->size.w-b.w)/2;
	b.y=(screen->size.h-b.h)/2;

	ht_dialog *d=new ht_dialog();
	d->init(&b, "clipboard", FS_TITLE | FS_KILLER);

	b.x=0;
	b.y=0;
	b.w-=2;
	b.h-=2;
	ht_clipboard_viewer *v=new ht_clipboard_viewer();
	v->init(&b, "clipboard", clipboard, VC_EDIT | VC_GOTO | VC_SEARCH);

	d->insert(v);
	d->setpalette(window_palette_str);

	execute_dialog(d);

	STREAMOFS s, e;
	v->select_get(&s, &e);
	if (e>s) {
		clipboard->clipboard_start=s;
		clipboard->clipboard_size=e-s;
	} else {
		clipboard->clipboard_start=0;
		clipboard->clipboard_size=0;
	}

	d->done();
	delete d;
}

void ht_app::popup_color_list()
{
	bounds b;
	getbounds(&b);
	b.x=b.w/6;
	b.y=b.h/6;
	b.w=2*b.w/3;
	b.h=2*b.h/3;
	ht_color_manager *dialog=new ht_color_manager();
	dialog->init(&b, pal_grp);
	if (execute_dialog(dialog)) {
		ht_color_manager_data data;
		dialog->getdata(&data);
		pal_grp->paste(data.pal_grp);
		sendmsg(htm_dirtyview, 0);
	}
	dialog->done();
	delete dialog;
}

#if 1
void ht_app::popup_view_list()
{
	bounds b, c;
	getbounds(&b);
	b.x=b.w/4;
	b.y=b.h/4;
	b.w/=2;
	b.h/=2;
	ht_dialog *dialog=new ht_dialog();
	dialog->init(&b, "views", FS_KILLER | FS_TITLE);

/* create listbox */
	c=b;
	c.x=0;
	c.y=0;
	c.w-=2;
	c.h-=2;
	ht_text_listbox *listbox=new ht_text_listbox();
	listbox->init(&c);
/* insert all browsable views */
	ht_collection *structure=new ht_collection();
	int index=0;
	popup_view_list_dump(battlefield->current, listbox, structure, 0, &index, battlefield->getselected());
/* and seek to the currently selected one */
	listbox->update();
	listbox->seek(index);

	dialog->insert(listbox);
	dialog->setpalette(windialog_special_palette_str);

	if (execute_dialog(dialog)) {
		ht_listbox_data data;
		listbox->getdata(&data);
		ht_data_ptr *p=(ht_data_ptr*)structure->get(((ht_text_listbox_item*)data.cursor)->id);
		focus((ht_view*)p->data);
	}
	dialog->done();
	delete dialog;
}

void ht_app::popup_view_list_dump(ht_view *view, ht_text_listbox *listbox, ht_collection *structure, int depth, int *currenti, ht_view *currentv)
{
	char str[256];
	char *s=str;
	if (!(view->options & VO_BROWSABLE)) {
		depth--;
	}

	for (int i=0; i<depth; i++) { *(s++)=' ';*(s++)=' '; }
	*s=0;

	if (!view) return;
	int c=view->childcount();
	for (int i=0; i<c; i++) {
		ht_view *v=view->getfirstchild();
		while (v) {
			if (v->browse_idx==i) break;
			v=v->next;
		}
		if (!v) return;
		if (v->options & VO_BROWSABLE) {
			sprintf(s, "- %s", v->desc);
			structure->insert(new ht_data_ptr(v));
			if (v==currentv)
				*currenti=structure->count()-1;
			listbox->insert_str(structure->count()-1, str);
		}
		popup_view_list_dump(v, listbox, structure, depth+1, currenti, currentv);
	}
}

#else

void ht_app::popup_view_list()
{
	bounds b, c;
	getbounds(&b);
	b.x=b.w/4;
	b.y=b.h/4;
	b.w/=2;
	b.h/=2;
	ht_dialog *dialog=new ht_dialog(&b, "views");

/* create listbox */
	c=b;
	c.x=0;
	c.y=0;
	c.w-=2;
	c.h-=2;
	ht_static_treeview *treeview=new ht_static_treeview(&c, "jo");
/* insert all browsable views */
//	ht_collection *structure=new ht_collection();
//	int index=0;
	popup_view_list_dump(treeview, current, 0, 0);
/* and seek to the currently selected one */
//	listbox->seek(index);

//	treeview->add_child(0, "ertg");

	dialog->insert(treeview);
	dialog->setpalette(windialog_special_palette_str);

	if (execute_dialog(dialog)) {
/*		ht_listbox_data data;
		listbox->getdata(&data);
		ht_data_ptr *p=(ht_data_ptr*)structure->get(data.index);
		focus((ht_view*)p->data);*/
	}
	dialog->done();
	delete dialog;
}

void ht_app::popup_view_list_dump(ht_static_treeview *treeview, ht_view *view, void *node, int depth)
{
	void *r;
	if ((view->options & VO_BROWSABLE) && (depth)) {
		char str[128];
		strcpy(str, view->desc);
		r=treeview->add_child(node, str);
	} else r=node;

	int c=view->childcount();
	for (int i=0; i<c; i++) {
		ht_view *v=view->getfirstchild();
		while (v) {
			if (v->browse_idx==i) break;
			v=v->next;
		}
		if (!v) return;
//		if (v->options & VO_BROWSABLE) {
			popup_view_list_dump(treeview, v, r, depth++);
//		}
	}
}
#endif


void ht_app::popup_window_list()
{
	bounds b, c;
	getbounds(&b);
	b.x=b.w/4;
	b.y=b.h/4;
	b.w/=2;
	b.h/=2;
	ht_dialog *dialog=new ht_dialog();
	dialog->init(&b, "windows", FS_KILLER | FS_TITLE);

/* create listbox */
	c=b;
	c.x=0;
	c.y=0;
	c.w-=2;
	c.h-=2;
	ht_text_listbox *listbox=new ht_text_listbox();
	listbox->init(&c);

	int view_count=0;
	ht_view *v=first;
	while (v) {
		view_count++;
		v=v->next;
	}

	for (int i=0; i<view_count; i++) {
		ht_view *v=battlefield->first;
		while (v) {
			if ((v->getnumber()==i) && (v->options & VO_BROWSABLE)) {
				char l[1024];	/* FIXME: possible buffer overflow ! */
				sprintf(l, "%2d %s", i, v->desc);
				listbox->insert_str(i, l);
			}
			v=v->next;
		}
	}
	listbox->update();
	listbox->seek(battlefield->current->getnumber()-1);

	dialog->insert(listbox);
	dialog->setpalette(windialog_special_palette_str);

	if (execute_dialog(dialog)) {
		ht_listbox_data data;
		listbox->getdata(&data);
		focus_window(((ht_text_listbox_item*)data.cursor)->id);
	}
	dialog->done();
	delete dialog;
}

/*
 *
 */

void jump_dialog(ht_viewer *viewer)
{
	char addr[1024];
	addr[0] = 0;
	if (inputbox("goto", "address/'symbol'", addr, 1024, HISTATOM_GOTO)) {
		viewer->jumpstr(addr);
	}
}

void search_dialog(ht_viewer *viewer, STREAMOFS current_ofs)
{
	bounds b;
	b.w=50;
	b.h=15;
	b.x=(screen->size.w-b.w)/2;
	b.y=(screen->size.h-b.h)/2;
	ht_search_dialog *dialog=new ht_search_dialog();
	dialog->init(&b);
	b.x=0;
	b.y=0;
	b.w-=2;
	b.h-=2;

	ht_hexascii_search_mask *s0=new ht_hexascii_search_mask();
	s0->init(&b, 0);
	dialog->insertmode(0, "ascii/hex", s0);

	ht_casm_search_mask *s1=new ht_casm_search_mask();
	s1->init(&b, 0);
	dialog->insertmode(1, "c/asm", s1);

	if (execute_dialog(dialog)) {
		ht_search_dialog_data t;
		dialog->getdata(&t);
		switch (t.id) {
			case 0: {
/* ascii/hex */
				ht_hexascii_search_mask_data *d=(ht_hexascii_search_mask_data*)malloc(t.view->datasize());
				t.view->getdata(d);
				if (d->strlen) {
					ht_search t;
					int error=0;

					t.type=ST_BIN;
					t.pattern_length=d->strlen;
					t.caseinsensitive=d->caseinsensitive;
					t.idle=d->idle;
					t.pattern=d->str;
					STREAMOFS start=current_ofs, end=0xffffffff;
					if ((!error) && (d->startlen)) {
						char str[64];
						memmove(str, d->start, 63);
						str[63]=0;
						VIEWERADDR a;
						if (!viewer->convert_str_to_addr(str, &a)) {
							error=1;
							errorbox("start_address: '%s' doesn't seem to be a valid address", &str);
						} else if (!viewer->convert_addr_to_ofs(a, &start)) {
							error=1;
							errorbox("start_address: '%s' doesn't seem to be a valid address", &str);
						}
					}
					if ((!error) && (d->endlen)) {
						char str[64];
						memmove(str, d->end, 63);
						str[63]=0;
						VIEWERADDR a;
						if (!viewer->convert_str_to_addr(str, &a)) {
							error=1;
							errorbox("end_address: '%s' doesn't seem to be a valid address", &str);
						} else if (!viewer->convert_addr_to_ofs(a, &end)) {
							error=1;
							errorbox("end_address: '%s' doesn't seem to be a valid address", &str);
						}
					}
					if (!error) {
//						infobox("debug: searching for '%s' from %08x to %08x", d->str, start, end);
						viewer->search(&t, start, end);
					}
				}
				delete d;
				break;
			}
			case 1: {
/* C/ASM */
				ht_casm_search_mask_data *d=(ht_casm_search_mask_data*)malloc(t.view->datasize());
				t.view->getdata(d);
				if (d->strlen) {
					char k[128];
					int l=d->strlen;
					if (l>127) l=127;
					memmove(k, d->str, l);
					k[l]=0;
					int slen;
					byte *s=eval_casmstr(k, &slen);
					if ((s) && (*s)) {
						ht_search t;
						int error=0;

						t.type=ST_BIN;
						t.pattern_length=slen;
						t.caseinsensitive=d->caseinsensitive;
						t.pattern=s;
						VIEWERADDR start=current_ofs, end=0xffffffff;
						if ((!error) && (d->startlen)) {
							char str[64];
							memmove(str, d->start, 63);
							str[63]=0;
							VIEWERADDR a;
							if (!viewer->convert_str_to_addr(str, &a)) {
								error=1;
								errorbox("start_address: '%s' doesn't seem to be a valid address", &str);
							} else if (!viewer->convert_addr_to_ofs(a, &start)) {
								error=1;
								errorbox("start_address: '%s' doesn't seem to be a valid address", &str);
							}
						}
						if ((!error) && (d->endlen)) {
							char str[64];
							memmove(str, d->end, 63);
							str[63]=0;
							VIEWERADDR a;
							if (!viewer->convert_str_to_addr(str, &a)) {
								error=1;
								errorbox("end_address: '%s' doesn't seem to be a valid address", &str);
							} else if (!viewer->convert_addr_to_ofs(a, &end)) {
								error=1;
								errorbox("end_address: '%s' doesn't seem to be a valid address", &str);
							}
						}
						if (!error) {
//							infobox("debug: searching for '%s' from %08x to %08x", d->str, start, end);
							viewer->search(&t, start, end);
						}
					} else errorbox("%s in string: %s", casmerrstr, k);
				}
				delete d;
				break;
			}
			default:
				warnbox("search mode (id=%d) not supported !", t.id);
		}
	}
	dialog->done();
	delete dialog;
}

/*
 *	INIT
 */

int init_app()
{
	palette *desktop_palette=pal_grp->create(desktop_palette_str, desktop_palette_colors, PCLASS_DESKTOP);
	pal_grp->set_entry(desktop_palette, desktop_coloridx_body, "body", VCP(VC_BLUE, VC_BLACK));

	palette *keyline_palette=pal_grp->create(keyline_palette_str, keyline_palette_colors, PCLASS_KEYLINE);
	pal_grp->set_entry(keyline_palette, keyline_coloridx_digit, "digit", VCP(VC_WHITE, VC_BLACK));
	pal_grp->set_entry(keyline_palette, keyline_coloridx_active, "active", VCP(VC_BLACK, VC_CYAN));
	pal_grp->set_entry(keyline_palette, keyline_coloridx_inactive, "inactive", VCP(VC_WHITE, VC_CYAN));

	bounds b;
	screen=new screendrawbuf();

	b.x=0;
	b.y=0;
	b.w=screen->size.w;
	b.h=screen->size.h;
	app=new ht_app();
	((ht_app*)app)->init(&b);
	baseview=app;

//     byte *spass=casm2str("\"\\0234535hallo\\n\\x1fea\\077\" #32 ,0xd 0x66 255, 0ah 15h 077o,100011b", 0);
//     byte *spass=casm2str("0, 0", 0);

	return 1;
}

/*
 *	DONE
 */

void done_app()
{
	app->done();
	delete app;
	delete screen;
}
